from typing import List, Union

import sympy as sp

from pystencils.astnodes import Node, ResolvedFieldAccess, Block
from pystencils.simp import AssignmentCollection
from pystencils.assignment import Assignment


# noinspection PyPep8Naming
class fast_division(sp.Function):
    """
    Produces special float instructions for CUDA kernels
    """
    nargs = (2,)


# noinspection PyPep8Naming
class fast_sqrt(sp.Function):
    """
    Produces special float instructions for CUDA kernels
    """
    nargs = (1, )


# noinspection PyPep8Naming
class fast_inv_sqrt(sp.Function):
    """
    Produces special float instructions for CUDA kernels
    """
    nargs = (1, )


class Fma(sp.Function):
    nargs = (3, )


# noinspection PyPep8Naming
class fmadd(Fma):
    """
    Produces FMA instructions for a*b+c
    """
    instruction = '*+'


# noinspection PyPep8Naming
class fmsub(Fma):
    """
    Produces FMA instructions for a*b-c
    """
    instruction = '*-'


# noinspection PyPep8Naming
class fnmadd(Fma):
    """
    Produces FMA instructions for -a*b+c
    """
    instruction = '-*+'


# noinspection PyPep8Naming
class fnmsub(Fma):
    """
    Produces FMA instructions for -a*b-c
    """
    instruction = '-*-'


def _run(term, visitor):
    if isinstance(term, AssignmentCollection):
        new_main_assignments = _run(term.main_assignments, visitor)
        new_subexpressions = _run(term.subexpressions, visitor)
        return term.copy(new_main_assignments, new_subexpressions)
    elif isinstance(term, list):
        return [_run(e, visitor) for e in term]
    else:
        return visitor(term)


def insert_fast_sqrts(term: Union[sp.Expr, List[sp.Expr], AssignmentCollection, Assignment]):
    def visit(expr):
        if isinstance(expr, Node):
            return expr
        if expr.func == sp.Pow and isinstance(expr.exp, sp.Rational) and expr.exp.q == 2:
            power = expr.exp.p
            if power < 0:
                return fast_inv_sqrt(expr.args[0]) ** (-power)
            else:
                return fast_sqrt(expr.args[0]) ** power
        else:
            new_args = [visit(a) for a in expr.args]
            return expr.func(*new_args) if new_args else expr
    return _run(term, visit)


def insert_fast_divisions(term: Union[sp.Expr, List[sp.Expr], AssignmentCollection, Assignment]):

    def visit(expr):
        if isinstance(expr, Node):
            return expr
        if expr.func == sp.Mul:
            div_args = []
            other_args = []
            for a in expr.args:
                if a.func == sp.Pow and a.exp.is_integer and a.exp < 0:
                    div_args.append(visit(a.base) ** (-a.exp))
                else:
                    other_args.append(visit(a))
            if div_args:
                return fast_division(sp.Mul(*other_args), sp.Mul(*div_args))
            else:
                return sp.Mul(*other_args)
        elif expr.func == sp.Pow and expr.exp.is_integer and expr.exp < 0:
            return fast_division(1, visit(expr.base) ** (-expr.exp))
        else:
            new_args = [visit(a) for a in expr.args]
            return expr.func(*new_args) if new_args else expr

    return _run(term, visit)


def insert_fma(term, operators):
    from pystencils.rng import RNGBase  # late import to avoid cyclic dependency

    if '*+' not in operators:
        return term

    def flatten(expr):
        if isinstance(expr, sp.Add):
            new_args = []
            for arg in expr.args:
                if arg.func == sp.Add:
                    new_args += [flatten(a) for a in arg.args]
                else:
                    new_args.append(flatten(arg))
            return sp.Add(*new_args)
        elif isinstance(expr, sp.Mul):
            new_args = []
            for arg in expr.args:
                if arg.func == sp.Mul:
                    new_args += [flatten(a) for a in arg.args]
                else:
                    new_args.append(flatten(arg))
            return sp.Mul(*new_args)
        return expr

    def visit(expr):
        # Special treatments for various types that cannot be reconstructed from their args
        if isinstance(expr, ResolvedFieldAccess):
            return expr
        elif isinstance(expr, RNGBase):
            return expr
        elif hasattr(expr, 'body'):
            old_parent = expr.body.parent if hasattr(expr.body, 'parent') else None
            expr.body = visit(expr.body)
            if old_parent is not None:
                expr.body.parent = old_parent
            return expr
        elif isinstance(expr, Block):
            return Block([visit(a) for a in expr.args])

        # Find patterns of Add and Mul nodes that can be fused
        if expr.func == sp.Add:
            expr = flatten(expr)
            summands = list(expr.args)
            if '-*+' in operators:
                for summand in expr.args:
                    if summand.func == sp.Mul and len(summand.args) >= 3 and -1 in summand.args:
                        summands.remove(summand)
                        factors = list(summand.args)
                        factors.remove(-1)
                        factors = [visit(f) for f in factors]
                        if not ('-*-' in operators and all(s.func == sp.Mul and -1 in s.args for s in summands)):
                            summands = [visit(s) for s in summands]
                            return sp.Add(fnmadd(factors[0], sp.Mul(*factors[1:]), summands[0]), *summands[1:])
            summands = list(expr.args)
            if '-*-' in operators:
                negative = [s for s in summands if s.func == sp.Mul and -1 in s.args]
                positive = []
                if len(negative) > 1:
                    positive = []
                    for summand in negative:
                        summands.remove(summand)
                        factors = list(summand.args)
                        factors.remove(-1)
                        positive.append(sp.Mul(*factors))
                for summand in positive:
                    if summand.func == sp.Mul and len(summand.args) >= 2:
                        positive.remove(summand)
                        factors = list(summand.args)
                        positive = [visit(s) for s in positive]
                        summands = [visit(f) for f in summands]
                        return sp.Add(fnmsub(factors[0], sp.Mul(*factors[1:]), sp.Add(*positive[:2])),
                                      -sp.Add(*positive[2:]), *summands)
            summands = list(expr.args)
            if '*-' in operators:
                for summand in summands:
                    if summand.func == sp.Mul and len(summand.args) >= 2 and -1 not in summand.args:
                        summands.remove(summand)
                        factors = list(summand.args)
                        for summand in summands:
                            if summand.func == sp.Mul and -1 in summand.args:
                                summands.remove(summand)
                                subfactors = list(summand.args)
                                subfactors.remove(-1)
                                factors = [visit(f) for f in factors]
                                summands = [visit(s) for s in summands]
                                subfactors = [visit(f) for f in subfactors]
                                return sp.Add(fmsub(factors[0], sp.Mul(*factors[1:]), sp.Mul(*subfactors)), *summands)
            summands = list(expr.args)
            for summand in summands:
                if summand.func == sp.Mul and len(summand.args) >= 2:
                    summands.remove(summand)
                    factors = list(summand.args)
                    factors = [visit(f) for f in factors]
                    summands = [visit(s) for s in summands]
                    return sp.Add(fmadd(factors[0], sp.Mul(*factors[1:]), summands[0]), *summands[1:])
            return expr
        # Find Mul with three factors, one of them -1, which can be fused
        elif expr.func == sp.Mul and -1 in expr.args:
            expr = flatten(expr)
            factors = list(expr.args)
            factors.remove(-1)
            factors = [visit(f) for f in factors]
            if '-*+' in operators:
                return fnmadd(factors[0], sp.Mul(*factors[1:]), 0)
            elif '-*-' in operators:
                return fnmsub(factors[0], sp.Mul(*factors[1:]), 0)
        new_args = [visit(a) for a in expr.args]
        return expr.func(*new_args) if new_args else expr

    return _run(term, visit)
